import json


def confluent_error_transform(err):
    pass


def kafka_python_error_transform(err):
    pass


def serializer(msg):
    return json.dumps(msg)


def deserializer(msg):
    return json.loads(msg)




import gevent
import confluent_kafka
from confluent_kafka.admin import AdminClient, NewPartitions
from confluent_kafka import KafkaException
from confluent_kafka import TIMESTAMP_LOG_APPEND_TIME
from confluent_kafka import TIMESTAMP_NOT_AVAILABLE
from confluent_kafka import TIMESTAMP_CREATE_TIME
from kafka_utils import log
from kafka_utils.kafka_exceptions import GetMetaDataFailException
from kafka_utils.kafka_exceptions import PartitionCreationException
from kafka_utils.kafka_exceptions import BufferFullException
from kafka_utils.kafka_exceptions import ProduceFailException
from kafka_utils.kafka_tools import confluent_error_transform
from kafka_utils.kafka_tools import serializer
from kafka_utils.kafka_tools import deserializer
from kafka_utils.kafka_msg import KafkaMsg


class ConfluentProducer(confluent_kafka.Producer):
    def __init__(self, conf):
        assert 'metadata.request.timeout.ms' in conf
        if conf['metadata.request.timeout.ms'] < 0:
            raise ValueError('metadata request timeout can\'t be negative')
        self._metadata_request_timeout = conf.pop('metadata.request.timeout.ms') / 1000
        confluent_kafka.Producer.__init__(self, conf)
        assert 'bootstrap.servers' in conf
        self._admin = AdminClient({'bootstrap.servers': conf['bootstrap.servers']})

    def flush(self, timeout=None):
        pending_msgs = confluent_kafka.Producer.flush(self, timeout=0)
        with gevent.Timeout(timeout, False):
            while pending_msgs > 0:
                gevent.sleep(0.1)
                pending_msgs = confluent_kafka.Producer.flush(self, timeout=0)
        return pending_msgs

    def create_enough_partition(self, topic, partition_num, timeout=None):
        try:
            cur_num = self.get_partition_num(topic)
        except GetMetaDataFailException as e:
            log.error('failed to get partition num from cluster metadata')
            raise PartitionCreationException from e
        if cur_num > partition_num:
            log.error('can\'t decrease the number of partitions')
            return
        elif cur_num == partition_num:
            log.info('no need to create partitions')
            return

        new_parts = [NewPartitions(topic, int(partition_num))]

        try:
            fs = self._admin.create_partitions(new_parts, validate_only=False)
        except KafkaException as e:
            log.error('partition creation failed')
            raise PartitionCreationException from e
        except TypeError or ValueError:
            log.error('partition creation failed due to invalid input')
            raise

        for topic, f in fs.items():
            try:
                f.result()
                self._wait_partition_creation(topic, partition_num, timeout)
            except GetMetaDataFailException as e:
                log.error('failed to get partition num from cluster metadata')
                raise PartitionCreationException from e
            except Exception as e:
                log.error('partition creation failed')
                raise PartitionCreationException from e

    def get_partition_num(self, topic):
        try:
            metadata = self.list_topics(topic, timeout=self._metadata_request_timeout)
            return len(metadata.topics[topic].partitions)
        except KafkaException as e:
            raise GetMetaDataFailException from e

    def get_existing_topics(self):
        try:
            metadata = self.list_topics(timeout=self._metadata_request_timeout)
        except KafkaException as e:
            raise GetMetaDataFailException from e
        topics = []
        for topic in metadata.topics:
            topics.append(topic)
        return topics

    def _wait_partition_creation(self, topic, target_num, timeout):
        with gevent.Timeout(timeout, False):
            while True:
                try:
                    num = self.get_partition_num(topic)
                except GetMetaDataFailException:
                    raise
                if num == target_num:
                    break
                else:
                    gevent.sleep(0.1)

    def send_msg(self, topic, key, msg, suc_cb=None, err_cb=None, partition=-1):
        assert isinstance(key, str)
        s_msg = serializer(msg)
        try:
            confluent_kafka.Producer.produce(self, topic, s_msg, key=key, partition=partition,
                                             on_delivery=self._make_delivery_callback(suc_cb, err_cb))
        except BufferError as e:
            raise BufferFullException from e
        except KafkaException as e:
            error = e.args[0]
            log.error('produce failed due to ' + error.str())
            raise ProduceFailException from e
        except NotImplementedError as e:
            log.error('produce failed due to lack of timestamp support')
            raise ProduceFailException from e

    def _make_delivery_callback(self, suc_cb, err_cb):
        def dummy_callback(_, _2):
            pass

        def delivery_callback(err, msg):
            if err is not None:
                err_cb(confluent_error_transform(err))
            elif msg.error() is not None:
                err_cb(confluent_error_transform(msg.error()))
            else:
                suc_cb(self._create_kafka_msg(msg))

        if suc_cb is None or err_cb is None:
            return dummy_callback
        return delivery_callback

    @staticmethod
    def _create_kafka_msg(confluent_msg):
        assert confluent_msg.error() is None
        topic = confluent_msg.topic()
        key = confluent_msg.key()
        value = deserializer(confluent_msg.value())
        timestamp = confluent_msg.timestamp()
        assert isinstance(timestamp, tuple) and len(timestamp) == 2
        if timestamp[0] == TIMESTAMP_NOT_AVAILABLE:
            timestamp = ('TIMESTAMP_NOT_IMPLEMENTED', None)
        elif timestamp[0] == TIMESTAMP_CREATE_TIME:
            timestamp = ('TIMESTAMP_CREATION_TIME', timestamp[1])
        else:
            assert timestamp[0] == TIMESTAMP_LOG_APPEND_TIME
            timestamp = ('TIMESTAMP_BROKER_RECEIVE_TIME', timestamp[1])
        partition = confluent_msg.partition()
        offset = confluent_msg.offset()
        return KafkaMsg(topic, key, value, timestamp, partition, offset)




from kafka_utils.confluent_producer import ConfluentProducer
from kafka_utils.kafka_python_producer import KafkaPythonProducer
from kafka_utils import MAX_PARTITION_NUMBER
from kafka_utils.consistent_partitioner import ConsistentPartitioner


class MsgSender(object):
    DEFAULT_CONFIG = {
        'brokers': 'localhost',
        'client.id': None,
        'partitions.num': None,
        'enable.consistent.partitioner': False,
        'topics': None,
        'support.utils': 'confluent',
        'required.acks': 1,
        'compression.type': 'None',
        'request.retries': 5,
        'request.retries.backoff.ms': 100,
        'request.timeout.ms': 300000,
        'buffering.max.memory': 1048576,
        'buffering.max.ms': 0,
        'batch.num.messages': 10000,
        'metadata.age.max.ms': 300000,
        'metadata.request.timeout.ms': 300000,
        'nagle.enbale': True,
        'reconnect.backoff.ms': 50,
        'reconnect.timeout.ms': 30000,
        'max.in.flight.requests.per.connection': 1000000,
        'exactly.once': False,
        'security.protocol': 'PLAINTEXT',
        'ssl.cafile': 'None',
        'ssl.certfile': 'None',
        'ssl.keyfile': 'None',
        'ssl.password': 'None',
        'sasl.mechanism': 'GSSAPI',
        'sasl.username': 'None',
        'sasl.password': 'None'
    }

    def __init__(self, configs):
        self._topics = list()
        self._partition_nums = list()
        self._producer_type = ConfluentProducer
        self._consistent_hash = False
        self._consistent_hasher = None
        configure = self._resolve_conf(configs)
        self._producer = self._producer_type(configure)
        self._create_partitions()
        if self._consistent_hash:
            self._consistent_hasher = self._create_consistent_partitioner()

    def _create_consistent_partitioner(self):
        for i in range(len(self._partition_nums)):
            if self._partition_nums[i] == 0:
                self._partition_nums[i] = self.get_partition_num_for_topic(self._topics[i])
        return ConsistentPartitioner(self._topics, self._partition_nums)

    def _resolve_conf(self, conf):
        if not isinstance(conf, dict):
            raise TypeError('configure must be of type dict')
        if 'brokers' not in conf:
            raise ValueError('the configuration must contains brokers')
        if 'support.utils' in conf:
            if conf.pop('support.utils') == 'kafka-python':
                self._producer_type = KafkaPythonProducer

        if 'enable.consistent.partitioner' in conf:
            if conf.pop('enable.consistent.partitioner'):
                self._consistent_hash = True

        if 'partitions.num' in conf:
            if not isinstance(conf['partitions.num'], list):
                raise TypeError('partitions must be of type list')
            if 'topics' not in conf:
                raise ValueError('topics must be specified with the partitions.num option')

        if 'topics' in conf:
            if not isinstance(conf['topics'], list):
                raise TypeError('topics must be of type list')
            self._topics = conf.pop('topics')
            if 'partitions.num' in conf:
                self._partition_nums = conf.pop('partitions.num')
                if len(self._partition_nums) != len(self._topics):
                    raise ValueError('topics must be of the same length as partitions.num')
            else:
                self._partition_nums = [0 for _ in range(len(self._topics))]

        if self._producer_type is ConfluentProducer:
            configure = self._resolve_confluent_conf(conf)
        elif self._producer_type is KafkaPythonProducer:
            configure = self._resolve_kafka_python_conf(conf)
        else:
            configure = None

        if len(conf) > 0:
            unrecognized_options = ''
            for item in conf:
                unrecognized_options += (str(item) + ', ')
            unrecognized_options = unrecognized_options[:len(unrecognized_options)-2]
            raise ValueError('unrecognized options: ' + unrecognized_options)

        return configure

    @staticmethod
    def _resolve_kafka_python_conf(conf):
        return None

    @staticmethod
    def _resolve_confluent_conf(conf):
        configure = dict()

        assert 'brokers' in conf
        configure['bootstrap.servers'] = conf.pop('brokers')
        configure['client.id'] = conf.pop(
            'client.id') if 'client.id' in conf \
            else MsgSender.DEFAULT_CONFIG['client.id']
        configure['acks'] = conf.pop(
            'required.acks') if 'required.acks' in conf \
            else MsgSender.DEFAULT_CONFIG['required.acks']
        configure['compression.codec'] = conf.pop(
            'compression.type') if 'compression.type' in conf \
            else MsgSender.DEFAULT_CONFIG['compression.type']
        configure['retries'] = conf.pop(
            'request.retries') if 'request.retries' in conf \
            else MsgSender.DEFAULT_CONFIG['request.retries']
        configure['retry.backoff.ms'] = conf.pop(
            'request.retries.backoff.ms') if 'request.retries.backoff.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['request.retries.backoff.ms']
        configure['delivery.timeout.ms'] = conf.pop(
            'request.timeout.ms') if 'request.timeout.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['request.timeout.ms']
        configure['queue.buffering.max.kbytes'] = conf.pop(
            'buffering.max.memory') if 'buffering.max.memory' in conf \
            else MsgSender.DEFAULT_CONFIG['buffering.max.memory']
        configure['queue.buffering.max.ms'] = conf.pop(
            'buffering.max.ms') if 'buffering.max.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['buffering.max.ms']
        configure['batch.num.messages'] = conf.pop(
            'batch.num.messages') if 'batch.num.messages' in conf \
            else MsgSender.DEFAULT_CONFIG['batch.num.messages']
        configure['metadata.max.age.ms'] = conf.pop(
            'metadata.age.max.ms') if 'metadata.age.max.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['metadata.age.max.ms']
        configure['socket.nagle.disable'] = True if 'nagle.enbale' in conf and not conf.pop(
            'nagle.enbale') else False
        configure['reconnect.backoff.ms'] = conf.pop(
            'reconnect.backoff.ms') if 'reconnect.backoff.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['reconnect.backoff.ms']
        configure['reconnect.backoff.max.ms'] = conf.pop(
            'reconnect.timeout.ms') if 'reconnect.timeout.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['reconnect.timeout.ms']
        configure['max.in.flight'] = conf.pop(
            'max.in.flight.requests.per.connection') if 'max.in.flight.requests.per.connection' in conf \
            else MsgSender.DEFAULT_CONFIG['max.in.flight.requests.per.connection']
        configure['security.protocol'] = conf.pop(
            'security.protocol') if 'security.protocol' in conf \
            else MsgSender.DEFAULT_CONFIG['security.protocol']
        configure['ssl.ca.location'] = conf.pop(
            'ssl.cafile') if 'ssl.cafile' in conf \
            else MsgSender.DEFAULT_CONFIG['ssl.cafile']
        configure['ssl.certificate.location'] = conf.pop(
            'ssl.certfile') if 'ssl.certfile' in conf \
            else MsgSender.DEFAULT_CONFIG['ssl.certfile']
        configure['ssl.key.location'] = conf.pop(
            'ssl.keyfile') if 'ssl.key.location' in conf \
            else MsgSender.DEFAULT_CONFIG['ssl.keyfile']
        configure['ssl.key.password'] = conf.pop(
            'ssl.password') if 'ssl.password' in conf \
            else MsgSender.DEFAULT_CONFIG['ssl.password']
        configure['sasl.mechanism'] = conf.pop(
            'sasl.mechanism') if 'sasl.mechanism' in conf \
            else MsgSender.DEFAULT_CONFIG['sasl.mechanism']
        configure['sasl.username'] = conf.pop(
            'sasl.username') if 'sasl.username' in conf \
            else MsgSender.DEFAULT_CONFIG['sasl.username']
        configure['sasl.password'] = conf.pop(
            'sasl.password') if 'sasl.password' in conf \
            else MsgSender.DEFAULT_CONFIG['sasl.password']
        configure['metadata.request.timeout.ms'] = conf.pop(
            'metadata.request.timeout.ms') if 'metadata.request.timeout.ms' in conf \
            else MsgSender.DEFAULT_CONFIG['metadata.request.timeout.ms']

        if 'exactly.once' in conf:
            if conf.pop('exactly.once'):
                configure['enable.idempotence'] = True
                configure['enable.gapless.guarantee'] = True
        return configure

    @staticmethod
    def _check_partition_nums(partition_num):
        if partition_num < 0 or partition_num > MAX_PARTITION_NUMBER:
            raise ValueError('invalid partition num')

    def _create_partitions(self):
        for i in range(len(self._topics)):
            self._check_partition_nums(self._partition_nums[i])
            if self._partition_nums[i] > 0:
                self._producer.create_enough_partition(self._topics[i], self._partition_nums[i])

    def send_sync(self, key, value, suc_cb=None, err_cb=None, timeout=None, topics=None):
        self.send_async(key, value, suc_cb, err_cb, topics)
        self._producer.flush(timeout)

    def send_async(self, key, value, suc_cb=None, err_cb=None, topics=None):
        if topics is None:
            topics = self._topics
        if not isinstance(key, str):
            raise TypeError('key must be of type str')
        for topic in topics:
            if self._consistent_hash:
                partition = self._consistent_hasher.get_key(topic, key)
                self._producer.send_msg(topic, key, value, suc_cb, err_cb, partition=partition)
            else:
                self._producer.send_msg(topic, key, value, suc_cb, err_cb)

    def flush(self, timeout=None):
        self._producer.flush(timeout)

    def set_topics(self, topics, partition_nums=None):
        if not isinstance(topics, list):
            raise TypeError('topics must be of type list')

        if partition_nums is not None:
            if not isinstance(partition_nums, list):
                raise TypeError('partition_nums must be of type list')
            if len(topics) != len(partition_nums):
                raise ValueError('topics must be of the same length as partition_nums')
        else:
            partition_nums = [0 for _ in range(len(topics))]

        self._topics = topics
        self._partition_nums = partition_nums
        self._create_partitions()
        if self._consistent_hasher:
            self._partitioner = self._create_consistent_partitioner()

    def get_partition_num_for_topic(self, topic):
        return self._producer.get_partition_num(topic)

    def get_existing_topics(self):
        return self._producer.get_existing_topics()

    def close(self):
        pass




class KafkaExceptions(Exception):
    pass


class PartitionCreationException(KafkaExceptions):
    pass


class BufferFullException(KafkaExceptions):
    pass


class ProduceFailException(KafkaExceptions):
    pass


class GetMetaDataFailException(KafkaExceptions):
    pass



from hash_ring import HashRing


class ConsistentPartitioner(object):
    def __init__(self, topics, partition_numbers):
        self._hashers = {}
        for i in range(len(topics)):
            keys = [key for key in range(partition_numbers[i])]
            self._hashers[topics[i]] = HashRing(keys)

    def get_key(self, topic, key):
        return self._hashers[topic].get_node(key)






from setuptools import setup, find_packages

setup(
    name='kafka-utils-netease',
    version='1.4.9',
    packages=find_packages(),
    author='liuweidong',
    author_email='discoheaven@163.com',
    description='Easy way to use Kafka in python',
    install_requires=['confluent-kafka',
                      'gevent',
                      'hash_ring']
)





class KafkaMsg:
    def __init__(self, topic, key, value, timestamp, partition, offset):
        self._topic = topic
        self._key = key
        self._value = value
        self._timestamp = timestamp
        self._partition = partition
        self._offset = offset

    def __len__(self):
        return len(self._value)

    @property
    def topic(self):
        return self._topic

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def partition(self):
        return self._partition

    @property
    def offset(self):
        return self._offset
